# Copyright (C) 2011 by Marek Aaron Sapota
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import print_function

from threading import Lock
import sqlite3, sys, traceback

def open(storage):
    return PersistentQueue(storage)

class PersistentQueue():
    def __init__(self, storage):
        self.lock = Lock()
        self.taken = set()
        self.storage = storage
        with sqlite3.connect(self.storage) as connection:
            c = connection.cursor()
            try:
                c.execute('''create table queue
                (id integer primary key, task text, data text)''')
                self.connection.commit()
            except:
                pass # table already created

    def fetch(self):
        with self.lock:
            with sqlite3.connect(self.storage) as connection:
                c = connection.cursor()
                data = c.execute('''select * from queue
                where id not in ({0}) limit 1'''.format(
                    str(list(self.taken))[1:-1])).fetchone()
                if data is not None:
                    self.taken.add(data[0])
                    try: # Python 2.x
                        data = (data[0], data[1].encode('utf-8'),
                                data[2].encode('utf-8'), )
                    except: # Python 3.x
                        pass
                return data

    def pop(self, key):
        with self.lock:
            with sqlite3.connect(self.storage) as connection:
                c = connection.cursor()
                c.execute('delete from queue where id = ?', (key, ))
                connection.commit()
                self.taken.remove(key)

    def release(self, key):
        with self.lock:
            self.taken.remove(key)

    def push(self, task, data):
        if sys.version_info.major < 3: # Python 2.x
            try:
                task = task.decode('utf-8')
                data = data.decode('utf-8')
            except:
                print('queue - Malformed message, discarding.\n{0}'.
                        format(traceback.format_exc()))
                return
        with self.lock:
            with sqlite3.connect(self.storage) as connection:
                c = connection.cursor()
                c.execute('insert into queue values (NULL, ?, ?)',
                        (task, data, ))
                connection.commit()
